## UNISIM_CHECK_SDL
## Checks if the SDL library is installed
## Takes one parameter: the main function name
#####################################################
AC_DEFUN([UNISIM_CHECK_SDL], [
    # Check if SDL path has been overloaded
    AC_ARG_WITH(sdl,
	AS_HELP_STRING([--with-sdl=<path>], [sdl library to use (will be completed with /include and /lib)]))

    if test -n "$with_sdl"; then
		AC_MSG_NOTICE([using SDL at $with_sdl])
		if test "${host}" = "${build}"; then
			AC_CHECK_PROG(sdl_config_installed, sdl-config, yes, no, $with_sdl/bin)
			if test "$sdl_config_installed" != "yes"; then
					AC_MSG_ERROR([sdl-config not found. Please install SDL development library.])
			fi
			AC_MSG_NOTICE([sdl-config found])
			sdl_cflags="`$with_sdl/bin/sdl-config --cflags`"
			sdl_libs="`$with_sdl/bin/sdl-config --libs`"
			AC_MSG_NOTICE([sdl-config says compiler needs option ${sdl_cflags} ${sdl_libs} to compile and link with SDL])
		else
			# cross-compilation detected
			case "${host}" in
				*mingw32*)
					sdl_cflags="-I${with_sdl}/include/SDL"
					sdl_libs="-L${with_sdl}/lib -lmingw32 -lSDLmain -lSDL -mwindows -lm -luser32 -lgdi32 -lwinmm"
					;;
				*)
					sdl_cflags="-I${with_sdl}/include/SDL"
					sdl_libs="-L${with_sdl}/lib -lSDL"
					;;
			esac
		fi
		CPPFLAGS=${CPPFLAGS}" ${sdl_cflags}"
		LIBS=${LIBS}" ${sdl_libs}"
    else
		AC_CHECK_PROG(sdl_config_installed, sdl-config, yes, no)
		if test "$sdl_config_installed" != "yes"; then
				AC_MSG_ERROR([sdl-config not found. Please install SDL development library.])
		fi
		AC_MSG_NOTICE([sdl-config found])
		sdl_cflags="`sdl-config --cflags`"
		sdl_libs="`sdl-config --libs`"
		AC_MSG_NOTICE([sdl-config says compiler needs option ${sdl_cflags} ${sdl_libs} to compile and link with SDL])
		CPPFLAGS=${CPPFLAGS}" ${sdl_cflags}"
		LIBS=${LIBS}" ${sdl_libs}"
	fi

    # Check for the main SDL header
    AC_CHECK_HEADER(SDL.h, broken_incsdl=no, broken_incsdl=yes)
	
    if test "$broken_incsdl" = "yes"; then
		AC_MSG_ERROR([SDL includes not found (SDL/*.hh). Please use --with-sdl=<path>])
    fi

    # Check for SDL functions
    UNISIM_CHECK_LIB(SDL, SDL_WaitThread, $1,
    UNISIM_CHECK_LIB(SDL, SDL_FreeSurface, $1,
    UNISIM_CHECK_LIB(SDL, SDL_DestroyMutex, $1,
    UNISIM_CHECK_LIB(SDL, SDL_Quit, $1,
    UNISIM_CHECK_LIB(SDL, SDL_Init, $1,
    UNISIM_CHECK_LIB(SDL, SDL_CreateMutex, $1,
    UNISIM_CHECK_LIB(SDL, SDL_CreateThread, $1,
    UNISIM_CHECK_LIB(SDL, SDL_mutexP, $1,
    UNISIM_CHECK_LIB(SDL, SDL_mutexV, $1,
    UNISIM_CHECK_LIB(SDL, SDL_GetKeyName, $1,
    UNISIM_CHECK_LIB(SDL, SDL_Delay, $1,
    UNISIM_CHECK_LIB(SDL, SDL_PollEvent, $1,
    UNISIM_CHECK_LIB(SDL, SDL_SetVideoMode, $1,
    UNISIM_CHECK_LIB(SDL, SDL_WM_SetCaption, $1,
    UNISIM_CHECK_LIB(SDL, SDL_CreateRGBSurface, $1,
    UNISIM_CHECK_LIB(SDL, SDL_LockSurface, $1,
    UNISIM_CHECK_LIB(SDL, SDL_UnlockSurface, $1,
    UNISIM_CHECK_LIB(SDL, SDL_UpperBlit, $1,
    UNISIM_CHECK_LIB(SDL, SDL_UpdateRect, $1,
    UNISIM_CHECK_LIB(SDL, SDL_SaveBMP_RW, $1, broken_sdl=no,
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes),
    broken_sdl=yes)

    if test "$broken_sdl" = "yes"; then
		AC_MSG_NOTICE([SDL not found. No video frame buffer or input devices will be available.])
    else
		AC_DEFINE([HAVE_SDL], [], [Whether SDL is available])
    fi
])
